module Occur where

open import Data.Unit.Base using (⊤; tt)
open import Data.Nat.Base
open import Data.Nat.Properties
open import Data.Nat.Properties.Simple
open import Data.Fin hiding (_+_; _≤_; _<_)
open import Data.Sum
open import Data.Product
open import Data.Maybe.Base
open import Data.Empty
open import Data.List
open import Function using (_∘_)
open import Relation.Nullary
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Regular
open import Properties

---------------------------------------------------------------

-- the contents of the same injection are equal
inj₁-equal : {A B : Set} → {a1 a2 : A} → inj₁ {B = B} a1 ≡ inj₁ a2 → a1 ≡ a2
inj₁-equal refl = refl

inj₂-equal : {A B : Set} → {b1 b2 : B} → inj₂ {A = A} b1 ≡ inj₂ b2 → b1 ≡ b2
inj₂-equal refl = refl

pair-equal : {A B : Set} → {a1 a2 : A} → {b1 b2 : B} → (a1 , b1) ≡ (a2 , b2) → a1 ≡ a2 × b1 ≡ b2
pair-equal refl = (refl , refl)

-- the contents of the same ⟨_⟩ are equal
⟨-⟩-equal : {F : Code} → {m : ℕ} → {d1 d2 : ⟦ F ⟧ (μ F m)} → ⟨_⟩ {F = F} d1 ≡ ⟨ d2 ⟩ → d1 ≡ d2
⟨-⟩-equal refl = refl

suc-equal : {m : ℕ} → {x y : Fin m} → Fin.suc x ≡ suc y → x ≡ y
suc-equal refl = refl

inject≤-equal : {m m′ : ℕ} → {x y : Fin m} → {m≤m′ : m ≤ m′} →
        inject≤ x m≤m′ ≡ inject≤ y m≤m′ → x ≡ y
inject≤-equal {x = zero} {zero} {s≤s m≤m′} eq = refl
inject≤-equal {x = zero} {suc y} {s≤s m≤m′} ()
inject≤-equal {x = suc x} {zero} {s≤s m≤m′} ()
inject≤-equal {x = suc x} {suc y} {s≤s m≤m′} eq =
  cong suc (inject≤-equal (suc-equal eq))

--

-- thin x y = y     (if y < x)
--          = suc y (if y >= x)
-- thin x y will never be x.
thin : {m : ℕ} → Fin (suc m) → Fin m → Fin (suc m)
thin {m} zero y = suc y
thin {suc m} (suc x) zero = zero
thin {suc m} (suc x) (suc y) = suc (thin x y)

-- thick x y = just y       (if y < x)
--           = nothing      (if y = x)
--           = just (y - 1) (if y > x)
thick : {m : ℕ} → (x y : Fin (suc m)) → Maybe (Fin m)
thick {m} zero zero = nothing -- x = y
thick {m} zero (suc y) = just y -- y > x
thick {zero} (suc ()) zero
thick {suc m} (suc x) zero = just zero -- y < x
thick {zero} (suc ()) (suc y)
thick {suc m} (suc x) (suc y) with thick {m} x y
... | just x′ = just (suc x′)
... | nothing = nothing -- x = y

thick₂ : {m : ℕ} → (x y : Fin (suc m)) →
        x ≡ y ⊎ Σ[ y′ ∈ Fin m ] thin x y′ ≡ y
thick₂ zero zero = inj₁ refl -- x = y
thick₂ zero (suc y) = inj₂ (y , refl) -- y > x
thick₂ {zero} (suc ()) zero
thick₂ {suc m} (suc x) zero = inj₂ (zero , refl) -- y < x
thick₂ {zero} (suc ()) (suc ())
thick₂ {suc m} (suc x) (suc y) with thick₂ x y
thick₂ {suc m} (suc x) (suc .x) | inj₁ refl = inj₁ refl -- x = y
thick₂ {suc m} (suc x) (suc .(thin x y′)) | inj₂ (y′ , refl) =
  inj₂ (suc y′ , refl)

-- thick x x is always nothing
thickxx≡nothing : ∀ {m : ℕ} (x : Fin (suc m)) → thick x x ≡ nothing
thickxx≡nothing zero = refl
thickxx≡nothing {zero} (suc ())
thickxx≡nothing {suc m} (suc x) with thickxx≡nothing x
... | a rewrite a = refl

-- thick₂ x x is always inj₁ refl
thick₂xx≡inj₁refl : ∀ {m : ℕ} (x : Fin (suc m)) → thick₂ x x ≡ inj₁ refl
thick₂xx≡inj₁refl zero = refl
thick₂xx≡inj₁refl {zero} (suc ())
thick₂xx≡inj₁refl {suc m} (suc x) with thick₂xx≡inj₁refl x
... | a rewrite a = refl

-- If thick x y is nothing, then x ≡ y
thickxy-x≡y : ∀ {m : ℕ} (x y : Fin (suc m)) → thick x y ≡ nothing → x ≡ y
thickxy-x≡y zero zero eq = refl
thickxy-x≡y zero (suc y) ()
thickxy-x≡y {zero} (suc ()) zero eq
thickxy-x≡y {suc m} (suc x) zero ()
thickxy-x≡y {zero} (suc ()) (suc y) eq
thickxy-x≡y {suc m} (suc x) (suc y) eq with thick x y | inspect (thick x) y
thickxy-x≡y {suc m} (suc x) (suc y) () | just x' | [ eq' ]
... | nothing | [ eq' ] = cong suc (thickxy-x≡y x y eq')

-- inject≤ preserves thick and thin
thickxyjust : {m : ℕ} → {x y : Fin (suc m)} → {y' : Fin m} → {n : ℕ} → (leq : m ≤ n) →
        thick x y ≡ just y' → thick (inject≤ x (s≤s leq)) (inject≤ y (s≤s leq)) ≡ just (inject≤ y' leq)
thickxyjust {x = zero} {zero} leq ()
thickxyjust {x = zero} {suc y} leq refl = refl
thickxyjust {zero} {suc x} {y} {()} leq eq
thickxyjust {suc m} {suc x} {zero} {n = zero} () refl
thickxyjust {suc m} {suc x} {zero} {n = suc n} (s≤s leq) refl = refl
thickxyjust {suc m} {suc x} {suc y} leq eq with thick x y | inspect (thick x) y
thickxyjust {suc m} {suc x} {suc y} leq () | nothing | _
thickxyjust {suc m} {suc x} {suc y} {y''} {zero} () eq | just y' | [ eq2 ]
thickxyjust {suc m} {suc x} {suc y} {zero} {suc n} (s≤s leq) () | just y' | [ eq2 ]
thickxyjust {suc m} {suc x} {suc y} {suc .y'} {suc n} (s≤s leq) refl | just y' | [ eq2 ]
  with thickxyjust {x = x} leq eq2
... | eq3 rewrite eq3 = refl

thin-max : {m : ℕ} → (y : Fin m) →
        thin (fromℕ m) y ≡ inject≤ y (n≤m+n 1 m)
thin-max {zero} ()
thin-max {suc m} zero = refl
thin-max {suc m} (suc y) = cong suc (thin-max y)

inject-thin : {m l : ℕ} →
        (x : Fin (suc m)) →
        (y′ : Fin m) →
        (m≤l : m ≤ l) →
        inject≤ (thin x y′) (s≤s m≤l) ≡
        thin (inject≤ x (s≤s m≤l)) (inject≤ y′ m≤l)
inject-thin zero y′ m≤l = refl
inject-thin {zero} (suc x) () m≤l
inject-thin {suc m} (suc x) zero (s≤s m≤l) = refl
inject-thin {suc m} (suc x) (suc y′) (s≤s m≤l) =
  cong suc (inject-thin x y′ m≤l)

inject-thin′ : {m l : ℕ} →
        (x : Fin (suc m)) →
        (y′ : Fin m) →
        (y : Fin (suc m)) →
        (m≤l : m ≤ l) →
        thin x y′ ≡ y →
        thin (inject≤ x (s≤s m≤l)) (inject≤ y′ m≤l) ≡
        inject≤ y (s≤s m≤l)
inject-thin′ x y′ y m≤l thinxy′≡y
  rewrite sym thinxy′≡y = sym (inject-thin x y′ m≤l)

inject-thick′ : {m : ℕ} → {x y : Fin (suc m)} →
        {y′ : Fin m} → {n : ℕ} → (leq : m ≤ n) →
        (thinxy′≡y : thin x y′ ≡ y) →
        thick₂ x y ≡ inj₂ (y′ , thinxy′≡y) →
        thick₂ (inject≤ x (s≤s leq)) (inject≤ y (s≤s leq)) ≡
        inj₂ (inject≤ y′ leq , inject-thin′ x y′ y leq thinxy′≡y)
inject-thick′ {x = zero} {zero} leq () ()
inject-thick′ {x = zero} {suc y} leq .refl refl
  with inject-thin′ zero y (suc y) leq refl
... | refl = refl
inject-thick′ {zero} {suc x} {y′ = ()} leq thinxy′≡y thick₂xy≡inj₂y′
inject-thick′ {suc m} {suc x} {zero} (s≤s leq) .refl refl = refl
inject-thick′ {suc m} {suc x} {suc y} (s≤s leq) thinxy′≡y thick₂xy≡inj₂y′
  with thick₂ x y | inspect (thick₂ x) y
inject-thick′ {suc m} {suc x} {suc .x} (s≤s leq) thinxy′≡y () | inj₁ refl | [ eq ]
inject-thick′ {suc m} {suc x} {suc .(thin x y′)} (s≤s leq) .refl refl | inj₂ (y′ , refl) | [ thick₂xy≡₂y′ ]
  with inject-thick′ leq refl thick₂xy≡₂y′
... | eq₁ with inject-thin x y′ leq
... | eq₂ rewrite eq₁ | eq₂ = refl

thick-inject-just : {l l₁ : ℕ} →
        (y : Fin (suc l)) →
        (x : Fin (suc l₁)) →
        (x′ : Fin l₁) →
        (l≤l₁ : l ≤ l₁) → 
        (l≤l₂ : l ≤ suc l₁) → 
        thick (inject≤ y (s≤s l≤l₁)) x ≡ just x′ →
        thick (inject≤ y (s≤s l≤l₂)) (inject≤ x (s≤s (n≤1+n l₁))) ≡
        just (inject≤ x′ (n≤1+n l₁))
thick-inject-just {l} {l₁} y x x′ l≤l₁ l≤l₂ eq
  rewrite sym (inject≤-trans y (s≤s l≤l₁) (s≤s (n≤1+n l₁)) (s≤s l≤l₂))
  = thickxyjust {x = inject≤ y (s≤s l≤l₁)} {y = x} (n≤1+n l₁) eq

thick-inject-nothing : {l l₁ : ℕ} →
        (y : Fin (suc l)) →
        (x : Fin (suc l₁)) →
        (l≤l₁ : l ≤ l₁) → 
        (l≤l₂ : l ≤ suc l₁) → 
        thick (inject≤ y (s≤s l≤l₁)) x ≡ nothing →
        thick (inject≤ y (s≤s l≤l₂)) (inject≤ x (s≤s (n≤1+n l₁))) ≡
        nothing
thick-inject-nothing {l} {l₁} y x l≤l₁ l≤l₂ eq
  rewrite sym (thickxy-x≡y (inject≤ y (s≤s l≤l₁)) x eq)
        | inject≤-trans y (s≤s l≤l₁) (s≤s (n≤1+n l₁)) (s≤s l≤l₂)
  = thickxx≡nothing (inject≤ y (s≤s l≤l₂))

-- thick to thin
thick-thin : {m : ℕ} → (x y : Fin (suc m)) → (y' : Fin m) →
        thick x y ≡ just y' → thin x y' ≡ y
thick-thin zero zero y' ()
thick-thin zero (suc y) .y refl = refl
thick-thin {zero} (suc ()) y y' eq
thick-thin {suc m} (suc x) zero .zero refl = refl
thick-thin {suc m} (suc x) (suc y) y' eq with thick {m} x y | inspect (thick {m} x) y
thick-thin {suc m} (suc x) (suc y) .(suc y') refl | just y' | [ eq' ] = cong suc (thick-thin x y y' eq')
thick-thin {suc m} (suc x) (suc y) y' () | nothing | [ eq' ]

private
  -- thin to thick
  thin-thick : {m : ℕ} → (x y : Fin (suc m)) → (y' : Fin m) →
          thin x y' ≡ y → thick x y ≡ just y'
  thin-thick zero zero y' ()
  thin-thick zero (suc y) .y refl = refl
  thin-thick {zero} (suc ()) y y' eq
  thin-thick {suc m} (suc x) .zero zero refl = refl
  thin-thick {suc m} (suc x) zero (suc y') ()
  thin-thick {suc m} (suc x) (suc y) (suc y') eq with thin x y' | inspect (thin x) y'
  thin-thick {suc m} (suc x) (suc y) (suc y') refl | .y | [ eq' ] with thin-thick x y y' eq'
  ... | thickxy≡justy' rewrite thickxy≡justy' = refl

-- thick is inverse of thin
thick-thin-eq : {m : ℕ} → (x : Fin (suc m)) → (y' : Fin m) →
          thick x (thin x y') ≡ just y'
thick-thin-eq x y' = thin-thick x (thin x y') y' refl

data Ctx : Set where
  Hole : Ctx -- hole
  _⊕l_ : (G₁ : Ctx) → (G₂ : Code) → Ctx -- left
  _⊕r_ : (G₁ : Code) → (G₂ : Ctx) → Ctx -- right
  _⊗l_ : (G₁ : Ctx) → (G₂ : Code) → Ctx -- left
  _⊗r_ : (G₁ : Code) → (G₂ : Ctx) → Ctx -- right

fill : (C : Ctx) → (F : Code) → Code
fill Hole F = F
fill (C₁ ⊕l G₂) F = fill C₁ (F ⊕ G₂)
fill (G₁ ⊕r C₂) F = fill C₂ (G₁ ⊕ F)
fill (C₁ ⊗l G₂) F = fill C₁ (F ⊗ G₂)
fill (G₁ ⊗r C₂) F = fill C₂ (G₁ ⊗ F)

Ctx2⟦_⟧ : (C : Ctx) → (R : Set) → Set
Ctx2⟦ Hole    ⟧ R = ⊤
Ctx2⟦ C₁ ⊕l G₂ ⟧ R = Ctx2⟦ C₁ ⟧ R
Ctx2⟦ G₁ ⊕r C₂ ⟧ R = Ctx2⟦ C₂ ⟧ R
Ctx2⟦ C₁ ⊗l G₂ ⟧ R = Ctx2⟦ C₁ ⟧ R × ⟦ G₂ ⟧ R
Ctx2⟦ G₁ ⊗r C₂ ⟧ R = ⟦ G₁ ⟧ R × Ctx2⟦ C₂ ⟧ R

-- generic context and location for zipper
-- usually, R = μ F m
Ctx⟦_⟧ : (G : Code) → (R : Set) → Set
Ctx⟦ U       ⟧ R = ⊥
Ctx⟦ I       ⟧ R = ⊤
Ctx⟦ G₁ ⊕ G₂ ⟧ R = Ctx⟦ G₁ ⟧ R ⊎ Ctx⟦ G₂ ⟧ R 
Ctx⟦ G₁ ⊗ G₂ ⟧ R = Ctx⟦ G₁ ⟧ R × ⟦ G₂ ⟧ R ⊎ ⟦ G₁ ⟧ R × Ctx⟦ G₂ ⟧ R

Context : (G : Code) → Set → Set
Context G R = List (Ctx⟦ G ⟧ R)

plug′ : (G : Code) → {m : ℕ} → {F : Code} →
       Ctx⟦ G ⟧ (μ F m) → μ F m → ⟦ G ⟧ (μ F m)
plug′ U ()
plug′ I tt t = t
plug′ (G₁ ⊕ G₂) (inj₁ f) t = inj₁ (plug′ G₁ f t)
plug′ (G₁ ⊕ G₂) (inj₂ f) t = inj₂ (plug′ G₂ f t)
plug′ (G₁ ⊗ G₂) (inj₁ (d₁ , f₂)) t = (plug′ G₁ d₁ t , f₂)
plug′ (G₁ ⊗ G₂) (inj₂ (f₁ , d₂)) t = (f₁ , plug′ G₂ d₂ t)

plug : {F : Code} → {m : ℕ} → Context F (μ F m) → μ F m → μ F m
plug {F} [] t = t
plug {F} (f ∷ fs) t = ⟨ plug′ F f (plug fs t) ⟩

plug-append : {F : Code} → {m : ℕ} →
        (fs₁ fs₂ : Context F (μ F m)) → (t : μ F m) →
        plug fs₁ (plug fs₂ t) ≡ plug (fs₁ ++ fs₂) t
plug-append [] fs₂ t = refl
plug-append {F} (f ∷ fs₁) fs₂ t =
  cong (λ t → ⟨ plug′ F f t ⟩) (plug-append fs₁ fs₂ t)

cons-not-nil : {A : Set} → (fs : List A) → (f : A) →
      fs ++ f ∷ [] ≡ [] → ⊥
cons-not-nil [] f ()
cons-not-nil (x ∷ fs) f ()

plugC : (C : Ctx) → (G : Code) → {m : ℕ} → {F : Code} →
        F ≡ fill C G → Ctx2⟦ C ⟧ (μ F m) →
        ⟦ G ⟧ (μ F m) → ⟦ F ⟧ (μ F m)
plugC Hole G refl tt d = d
plugC (C₁ ⊕l G₂) G eq c d = plugC C₁ (G ⊕ G₂) eq c (inj₁ d)
plugC (G₁ ⊕r C₂) G eq c d = plugC C₂ (G₁ ⊕ G) eq c (inj₂ d)
plugC (C₁ ⊗l G₂) G eq (c₁ , d₂) d = plugC C₁ (G ⊗ G₂) eq c₁ (d , d₂)
plugC (G₁ ⊗r C₂) G eq (d₁ , c₂) d = plugC C₂ (G₁ ⊗ G) eq c₂ (d₁ , d)

plugC-plug : (C : Ctx) → {F : Code} → (G : Code) → {m : ℕ} →
          (eq : F ≡ fill C G) →
          (c : Ctx2⟦ C ⟧ (μ F m)) →
          (f : Ctx⟦ G ⟧ (μ F m)) →
          (t : μ F m) →
          Σ[ f′ ∈ Ctx⟦ F ⟧ (μ F m) ]
          plugC C G eq c (plug′ G f t) ≡ plug′ F f′ t
plugC-plug Hole G refl tt f t = (f , refl)
plugC-plug (C₁ ⊕l G₂) G eq c f t = plugC-plug C₁ (G ⊕ G₂) eq c (inj₁ f) t
plugC-plug (G₁ ⊕r C₂) G eq c f t = plugC-plug C₂ (G₁ ⊕ G) eq c (inj₂ f) t
plugC-plug (C₁ ⊗l G₂) G eq (c₁ , d₂) f t =
  plugC-plug C₁ (G ⊗ G₂) eq c₁ (inj₁ (f , d₂)) t
plugC-plug (G₁ ⊗r C₂) G eq (d₁ , c₂) f t =
  plugC-plug C₂ (G₁ ⊗ G) eq c₂ (inj₂ (d₁ , f)) t

private
  ⟨check⟩′ : {F : Code} → (G : Code) → {m : ℕ} → (x : Fin (suc m)) →
          (d : ⟦ G ⟧ (μ F (suc m))) →
          (r : ⟦ G ⟧′ (λ t → 
            (Σ[ fs ∈ Context F (μ F (suc m)) ]
             (plug fs ⟪ x ⟫ ≡ t))
          ⊎ (Σ[ t′ ∈ μ F m ]
             (⟪cata⟫ (⟪_⟫ ∘ thin x) t′ ≡ t))) d) →
          (Σ[ f ∈ Ctx⟦ G ⟧ (μ F (suc m)) ]
           Σ[ fs ∈ Context F (μ F (suc m)) ]
           (plug′ G f (plug fs ⟪ x ⟫) ≡ d))
        ⊎ (Σ[ d′ ∈ ⟦ G ⟧ (μ F m) ]
           (fmap G ⟨_⟩ (⟪_⟫ ∘ thin x) d′ ≡ d))
  ⟨check⟩′ U x tt tt = inj₂ (tt , refl)
  ⟨check⟩′ I x d (inj₁ (fs , eq)) = inj₁ (tt , fs , eq)
  ⟨check⟩′ I x d (inj₂ (d₁ , eq)) = inj₂ (d₁ , eq)
  ⟨check⟩′ (G₁ ⊕ G₂) x (inj₁ d) r with ⟨check⟩′ G₁ x d r
  ... | inj₁ (f , fs , eq) = inj₁ (inj₁ f , fs , cong inj₁ eq)
  ... | inj₂ (d′ , eq) = inj₂ (inj₁ d′ , cong inj₁ eq)
  ⟨check⟩′ (G₁ ⊕ G₂) x (inj₂ d) r with ⟨check⟩′ G₂ x d r
  ... | inj₁ (f , fs , eq) = inj₁ (inj₂ f , fs , cong inj₂ eq)
  ... | inj₂ (d′ , eq) = inj₂ (inj₂ d′ , cong inj₂ eq)
  ⟨check⟩′ (G₁ ⊗ G₂) x (d₁ , d₂) (r₁ , r₂) with ⟨check⟩′ G₁ x d₁ r₁
  ... | inj₁ (f₁ , fs₁ , eq₁) = inj₁ (inj₁ (f₁ , d₂) , fs₁ , cong (λ z → (z , d₂)) eq₁)
  ... | inj₂ (d₁′ , eq₁)
    with ⟨check⟩′ G₂ x d₂ r₂
  ... | inj₁ (f₂ , fs₂ , eq₂) = inj₁ (inj₂ (d₁ , f₂) , fs₂ , cong (λ z → (d₁ , z)) eq₂)
  ... | inj₂ (d₂′ , eq₂) = inj₂ ((d₁′ , d₂′) , cong₂ _,_ eq₁ eq₂)

  ⟨check⟩ : (F : Code) → {m : ℕ} → (x : Fin (suc m)) →
          (d : ⟦ F ⟧ (μ F (suc m))) →
          (r : ⟦ F ⟧′ (λ t → 
            (Σ[ fs ∈ Context F (μ F (suc m)) ]
             (plug fs ⟪ x ⟫ ≡ t))
          ⊎ (Σ[ t′ ∈ μ F m ]
             (cata ⟨_⟩ (⟪_⟫ ∘ thin x) t′ ≡ t))) d) →
          (Σ[ fs ∈ Context F (μ F (suc m)) ]
           (plug {F} fs ⟪ x ⟫ ≡ ⟨ d ⟩))
        ⊎ (Σ[ t′ ∈ μ F m ]
           (cata (⟨_⟩ {F}) (⟪_⟫ ∘ thin x) t′ ≡ ⟨ d ⟩))
  ⟨check⟩ F x d r with ⟨check⟩′ F x d r
  ... | inj₁ (f , fs , eq) = inj₁ (f ∷ fs , cong ⟨_⟩ eq)
  ... | inj₂ (d′ , eq) = inj₂ (⟨ d′ ⟩ , cong ⟨_⟩ eq)

-- private
⟪check⟫ : {F : Code} → {m : ℕ} → (x y : Fin (suc m)) →
        (Σ[ C ∈ Context F (μ F (suc m)) ] plug {F} C ⟪ x ⟫ ≡ ⟪ y ⟫)
      ⊎ (Σ[ t′ ∈ μ F m ] ⟪cata⟫ (⟪_⟫ ∘ thin x) t′ ≡ ⟪ y ⟫)
⟪check⟫ x y with thick₂ x y
... | inj₁ x≡y = inj₁ ([] , (cong ⟪_⟫ x≡y)) -- x occurs in y
... | inj₂ (y′ , thinxy′≡y) = inj₂ (⟪ y′ ⟫ , (cong ⟪_⟫ thinxy′≡y))

-- check x t: check whether x occurs in t with correctness proof
check : {F : Code} → {m : ℕ} →
         (x : Fin (suc m)) → (t : μ F (suc m)) →
         (Σ[ C ∈ Context F (μ F (suc m)) ] plug C ⟪ x ⟫ ≡ t)
       ⊎ (Σ[ t′ ∈ μ F m ] ⟪cata⟫ (⟪_⟫ ∘ thin x) t′ ≡ t)
check {F} {m} x t =
  ind (λ t → (Σ[ fs ∈ Context F (μ F (suc m)) ] plug fs ⟪ x ⟫ ≡ t)
             ⊎ (Σ[ t′ ∈ μ F m ] cata ⟨_⟩ (⟪_⟫ ∘ thin x) t′ ≡ t))
      (⟨check⟩ F x) (⟪check⟫ x) t

private
  ⟨noCycle⟩′ : (C : Ctx) → {F : Code} → (G : Code) → {m : ℕ} →
          (eq : F ≡ fill C G) →
          (d : ⟦ G ⟧ (μ F m)) →
          (r : ⟦ G ⟧′ (λ t → (c : Context F (μ F m)) →
                              plug c t ≡ t → c ≡ []) d) →
          (c : Ctx2⟦ C ⟧ (μ F m)) →
          (f : Ctx⟦ G ⟧ (μ F m)) →
          (fs : Context F (μ F m)) →
          plug′ G f (plug fs ⟨ plugC C G eq c d ⟩) ≡ d → ⊥
  ⟨noCycle⟩′ C U eq tt tt c ()
  ⟨noCycle⟩′ C I eq d r c tt fs eq′ with plugC-plug C I eq c tt d
  ... | (f , eq′′)
    rewrite eq′′ | plug-append fs (f ∷ []) d
    with r (fs ++ f ∷ []) eq′
  ... | eq′′′ = cons-not-nil fs f eq′′′
  ⟨noCycle⟩′ C (G₁ ⊕ G₂) eq (inj₁ d) r c (inj₁ f) fs eq′ =
    ⟨noCycle⟩′ (C ⊕l G₂) G₁ eq d r c f fs (inj₁-equal eq′)
  ⟨noCycle⟩′ C (G₁ ⊕ G₂) eq (inj₁ d) r c (inj₂ f) fs ()
  ⟨noCycle⟩′ C (G₁ ⊕ G₂) eq (inj₂ d) r c (inj₁ f) fs ()
  ⟨noCycle⟩′ C (G₁ ⊕ G₂) eq (inj₂ d) r c (inj₂ f) fs eq′ =
    ⟨noCycle⟩′ (G₁ ⊕r C) G₂ eq d r c f fs (inj₂-equal eq′)
  ⟨noCycle⟩′ C (G₁ ⊗ G₂) eq (d₁ , d₂) (r₁ , r₂) c (inj₁ (f , d)) fs eq′ =
    ⟨noCycle⟩′ (C ⊗l G₂) G₁ eq d₁ r₁ (c , d₂) f fs (proj₁ (pair-equal eq′))
  ⟨noCycle⟩′ C (G₁ ⊗ G₂) eq (d₁ , d₂) (r₁ , r₂) c (inj₂ (d , f)) fs eq′ =
    ⟨noCycle⟩′ (G₁ ⊗r C) G₂ eq d₂ r₂ (d₁ , c) f fs (proj₂ (pair-equal eq′))

  ⟨noCycle⟩ : (F : Code) → {m : ℕ} →
          (d : ⟦ F ⟧ (μ F m)) →
          (r : ⟦ F ⟧′ (λ t → (c : Context F (μ F m)) →
                              plug c t ≡ t → c ≡ []) d) →
          (c : Context F (μ F m)) →
          plug {F} c ⟨ d ⟩ ≡ ⟨ d ⟩ → c ≡ []
  ⟨noCycle⟩ F d r [] refl = refl
  ⟨noCycle⟩ F d r (f ∷ fs) eq =
    ⊥-elim (⟨noCycle⟩′ Hole F refl d r tt f fs (⟨-⟩-equal eq))

  ⟪noCycle⟫ : (F : Code) → {m : ℕ} → (x : Fin m) →
          (c : Context F (μ F m)) →
          plug {F} c ⟪ x ⟫ ≡ ⟪ x ⟫ → c ≡ []
  ⟪noCycle⟫ F x [] eq = refl
  ⟪noCycle⟫ F x (f ∷ fs) ()

noCycle : {F : Code} → {m : ℕ} →
          (t : μ F m) →
          (c : Context F (μ F m)) →
          plug c t ≡ t → c ≡ []
noCycle {F} {m} t =
  ind (λ t → (c : Context F (μ F m)) →
                 plug c t ≡ t → c ≡ [])
      (λ d r c eq → ⟨noCycle⟩ F d r c eq)
      (λ x c eq → ⟪noCycle⟫ F x c eq)
      t
