module Regular where

open import Data.Unit.Base using (⊤; tt)
open import Data.Nat.Base
open import Data.Nat.Properties using (n≤m+n)
open import Data.Fin hiding (_≤_; _+_)
open import Data.Sum
open import Data.Product
open import Relation.Binary.PropositionalEquality

---------------------------------------------------------------

infixr 3 _⊕_
infixr 4 _⊗_

data Code : Set where
  U   : Code -- unit
  I   : Code -- recursive position
  _⊕_ : (F G : Code) → Code -- sum
  _⊗_ : (F G : Code) → Code -- product

⟦_⟧ : (F : Code) → (A : Set) → Set
⟦ U     ⟧ A = ⊤
⟦ I     ⟧ A = A
⟦ F ⊕ G ⟧ A = ⟦ F ⟧ A ⊎ ⟦ G ⟧ A
⟦ F ⊗ G ⟧ A = ⟦ F ⟧ A × ⟦ G ⟧ A

⟦_⟧′ : (F : Code) → {R : Set} → (P : R → Set) → (d : ⟦ F ⟧ R) → Set
⟦ U     ⟧′ P tt = ⊤
⟦ I     ⟧′ P d = P d
⟦ F ⊕ G ⟧′ P (inj₁ d) = ⟦ F ⟧′ P d
⟦ F ⊕ G ⟧′ P (inj₂ d) = ⟦ G ⟧′ P d
⟦ F ⊗ G ⟧′ P (d₁ , d₂) = ⟦ F ⟧′ P d₁ × ⟦ G ⟧′ P d₂

-- generic data type with m metavariables
data μ (F : Code) (m : ℕ) : Set where
  ⟨_⟩ : ⟦ F ⟧ (μ F m) → μ F m -- fixed point
  ⟪_⟫ : (x : Fin m) → μ F m -- metavariable

private
  -- original fmap (not used)
  fmap′ : (G : Code) → {A B : Set} →
         (f : A → B) → (⟦ G ⟧ A → ⟦ G ⟧ B)
  fmap′ U       f tt        = tt
  fmap′ I       f d         = f d -- apply f to recursive position
  fmap′ (G₁ ⊕ G₂) f (inj₁ d) = inj₁ (fmap′ G₁ f d)
  fmap′ (G₁ ⊕ G₂) f (inj₂ d) = inj₂ (fmap′ G₂ f d)
  fmap′ (G₁ ⊗ G₂) f (d₁ , d₂) = (fmap′ G₁ f d₁ , fmap′ G₂ f d₂)

  -- catamorphism (fold)

  -- original cata
  {-# NO_TERMINATION_CHECK #-}
  cata′ : {F : Code} → {m : ℕ} → {A : Set} →
          (⟨f⟩ : ⟦ F ⟧ A → A) → (⟪f⟫ : Fin m → A) → (μ F m → A)
  cata′ {F} ⟨f⟩ ⟪f⟫ ⟨ d ⟩ = ⟨f⟩ (fmap′ F (cata′ ⟨f⟩ ⟪f⟫) d)
  cata′ {F} ⟨f⟩ ⟪f⟫ ⟪ x ⟫ = ⟪f⟫ x

  -- fmap′ specialized to f = cata F ⟨f⟩ ⟪f⟫
mutual
  fmap : {F : Code} → (G : Code) → {m : ℕ} → {A : Set} →
         (⟨f⟩ : ⟦ F ⟧ A → A) → (⟪f⟫ : Fin m → A) →
         (⟦ G ⟧ (μ F m) → ⟦ G ⟧ A)
  fmap U        ⟨f⟩ ⟪f⟫ tt       = tt
  fmap I        ⟨f⟩ ⟪f⟫ d        = cata ⟨f⟩ ⟪f⟫ d -- apply cata to rec. pos.
  fmap (G₁ ⊕ G₂) ⟨f⟩ ⟪f⟫ (inj₁ d) = inj₁ (fmap G₁ ⟨f⟩ ⟪f⟫ d)
  fmap (G₁ ⊕ G₂) ⟨f⟩ ⟪f⟫ (inj₂ d) = inj₂ (fmap G₂ ⟨f⟩ ⟪f⟫ d)
  fmap (G₁ ⊗ G₂) ⟨f⟩ ⟪f⟫ (d₁ , d₂) =
    (fmap G₁ ⟨f⟩ ⟪f⟫ d₁ , fmap G₂ ⟨f⟩ ⟪f⟫ d₂)

  cata : {F : Code} → {m : ℕ} → {A : Set} →
          (⟨f⟩ : ⟦ F ⟧ A → A) → (⟪f⟫ : Fin m → A) → (μ F m → A)
  cata {F} ⟨f⟩ ⟪f⟫ ⟨ d ⟩ = ⟨f⟩ (fmap F ⟨f⟩ ⟪f⟫ d)
  cata {F} ⟨f⟩ ⟪f⟫ ⟪ x ⟫ = ⟪f⟫ x
  -- cata that uses specialized fmap by passing ⟨f⟩ and ⟪f⟫ instead of cata F ⟨f⟩ ⟪f⟫

private
  -- original everywhere (not used)
  everywhere′ : (G : Code) → {R : Set} → (P : R → Set) →
               (f : (t : R) → P t) → (d : ⟦ G ⟧ R) → ⟦ G ⟧′ P d
  everywhere′ U P f tt = tt
  everywhere′ I P f d = f d -- apply f to recursive position
  everywhere′ (G₁ ⊕ G₂) P f (inj₁ d) = everywhere′ G₁ P f d
  everywhere′ (G₁ ⊕ G₂) P f (inj₂ d) = everywhere′ G₂ P f d
  everywhere′ (G₁ ⊗ G₂) P f (d₁ , d₂) =
    (everywhere′ G₁ P f d₁ , everywhere′ G₂ P f d₂)

  -- induction principle

  -- original ind
  {-# NO_TERMINATION_CHECK #-}
  ind′ : {m : ℕ} → (F : Code) → (P : μ F m → Set) →
          (⟨f⟩ : (d : ⟦ F ⟧ (μ F m)) → ⟦ F ⟧′ P d → P ⟨ d ⟩) →
          (⟪f⟫ : (x : Fin m) → P ⟪ x ⟫) → (t : μ F m) → P t
  ind′ F P ⟨f⟩ ⟪f⟫ ⟨ d ⟩ = ⟨f⟩ d (everywhere′ F P (ind′ F P ⟨f⟩ ⟪f⟫) d)
  ind′ F P ⟨f⟩ ⟪f⟫ ⟪ x ⟫ = ⟪f⟫ x

  -- everywhere specialized to f = 
mutual
  everywhere : {F : Code} → (G : Code) → {m : ℕ} → (P : μ F m → Set) →
               (⟨f⟩ : (d : ⟦ F ⟧ (μ F m)) → ⟦ F ⟧′ P d → P ⟨ d ⟩) →
               (⟪f⟫ : (x : Fin m) → P ⟪ x ⟫) →
               (d : ⟦ G ⟧ (μ F m)) → ⟦ G ⟧′ P d
  everywhere U P ⟨f⟩ ⟪f⟫ tt = tt
  everywhere I P ⟨f⟩ ⟪f⟫ d = ind P ⟨f⟩ ⟪f⟫ d -- apply ind to rec. pos.
  everywhere (G₁ ⊕ G₂) P ⟨f⟩ ⟪f⟫ (inj₁ d) = everywhere G₁ P ⟨f⟩ ⟪f⟫ d
  everywhere (G₁ ⊕ G₂) P ⟨f⟩ ⟪f⟫ (inj₂ d) = everywhere G₂ P ⟨f⟩ ⟪f⟫ d
  everywhere (G₁ ⊗ G₂) P ⟨f⟩ ⟪f⟫ (d₁ , d₂) =
    (everywhere G₁ P ⟨f⟩ ⟪f⟫ d₁ , everywhere G₂ P ⟨f⟩ ⟪f⟫ d₂)

  ind : {F : Code} → {m : ℕ} → (P : μ F m → Set) →
          (⟨f⟩ : (d : ⟦ F ⟧ (μ F m)) → ⟦ F ⟧′ P d → P ⟨ d ⟩) →
          (⟪f⟫ : (x : Fin m) → P ⟪ x ⟫) → (t : μ F m) → P t
  ind {F} P ⟨f⟩ ⟪f⟫ ⟨ d ⟩ = ⟨f⟩ d (everywhere F P ⟨f⟩ ⟪f⟫ d)
  ind {F} P ⟨f⟩ ⟪f⟫ ⟪ x ⟫ = ⟪f⟫ x
  -- ind that uses specialized everywhere by passing ⟨f⟩ and ⟪f⟫ instead of cata F ⟨f⟩ ⟪f⟫

-- map on meta-variables
⟪fmap⟫ : {F : Code} → (G : Code) → {m m′ : ℕ} →
         (⟪f⟫ : Fin m → μ F m′) → (d : ⟦ G ⟧ (μ F m)) → ⟦ G ⟧ (μ F m′)
⟪fmap⟫ G ⟪f⟫ d = fmap G ⟨_⟩ ⟪f⟫ d

⟪cata⟫ : {F : Code} → {m m′ : ℕ} →
        (⟪f⟫ : Fin m → μ F m′) → (μ F m → μ F m′)
⟪cata⟫ ⟪f⟫ t = cata ⟨_⟩ ⟪f⟫ t

-- turn x into a meta-variable after lifting
⟪inject≤⟫ : {F : Code} → {m n : ℕ} →
        (m≤n : m ≤ n) → (x : Fin m) → μ F n
⟪inject≤⟫ m≤n x = ⟪ inject≤ x m≤n ⟫

-- lifting number of meta-variables
lift≤ : {F : Code} → {m m′ : ℕ} → m ≤ m′ → μ F m → μ F m′
lift≤ m≤m′ t = ⟪cata⟫ (λ x → ⟪ inject≤ x m≤m′ ⟫) t
-- lift≤ m≤m′ t = cata ⟨_⟩ (⟪inject≤⟫ m≤m′) t

-- map on meta-variables with lifting
⟪fmap⟫≤ : {F : Code} → (G : Code) → {m m′ m′′ : ℕ} →
        (⟪f⟫ : Fin m′′ → μ F m′) → (m≤m′′ : m ≤ m′′) →
        (d : ⟦ G ⟧ (μ F m)) → ⟦ G ⟧ (μ F m′)
⟪fmap⟫≤ G ⟪f⟫ m≤m′′ d =
  ⟪fmap⟫ G ⟪f⟫ (⟪fmap⟫ G (λ x → ⟪ inject≤ x m≤m′′ ⟫) d)

⟪cata⟫≤ : {F : Code} → {m m′ m′′ : ℕ} →
        (⟪f⟫ : Fin m′′ → μ F m′) → (m≤m′′ : m ≤ m′′) →
        (t : μ F m) → μ F m′
⟪cata⟫≤ ⟪f⟫ m≤m′′ t = ⟪cata⟫ ⟪f⟫ (lift≤ m≤m′′ t)
